#include <cxxtest/TestSuite.h>

#include "Event.h"
#include "RealMalloc.h"

class MockForEvent :
    public RealMalloc
{
public:
};

class EventTestSuite : public CxxTest::TestSuite
{
public:
	MockForEvent *mock;

	void setUp()
	{
		TS_ASSERT( mock = new MockForEvent );
	}

	void tearDown()
	{
		TS_ASSERT_THROWS_NOTHING( delete mock );
	}

    void test_type_and_equality()
    {
        Event event1( 1U );
        Event event2( 2U );
        Event event3( 1U );

        char payload[] = { 0x01, 0x02, 0x03 };
        Event event4( 2U, ConstSimpleData( payload, sizeof(payload) ) );
        Event event5( 1U, ConstSimpleData( payload, sizeof(payload) ) );

        TS_ASSERT_EQUALS( event1.type(), 1U );
        TS_ASSERT_EQUALS( event2.type(), 2U );
        TS_ASSERT_EQUALS( event3.type(), 1U );
        TS_ASSERT_EQUALS( event4.type(), 2U );
        TS_ASSERT_EQUALS( event5.type(), 1U );

        TS_ASSERT( event1 == event1 );
        TS_ASSERT( event3 == event5 );
        TS_ASSERT( event5 == event1 );

        TS_ASSERT( !( event2 == event1 ) );
        TS_ASSERT( !( event2 == event3 ) );
        TS_ASSERT( !( event2 == event5 ) );

        TS_ASSERT( event2 == event4 );

        TS_ASSERT( !(event1 != event1 ) );
        TS_ASSERT( !(event1 != event3 ) );
        TS_ASSERT( event2 != event1 );
        TS_ASSERT( event2 != event3 );

        Event event6;
        TS_ASSERT_EQUALS( event6.type(), 0xffffffff );
    }

    void test_serialize__no_payload()
    {
        Event event1( 0x1a2b3c4d );

        HeapData serialized;
        event1.serialize( serialized );

        char expectedSeriazlied[] = {
            0x4d, 0x3c, 0x2b, 0x1a,
            0x00, 0x00, 0x00, 0x00 };

        TS_ASSERT_EQUALS( serialized.size(), sizeof(expectedSeriazlied) );
        TS_ASSERT_SAME_DATA( serialized.data(), expectedSeriazlied, sizeof(expectedSeriazlied) );
    }

    void test_serialize__with_payload()
    {
        char data[] = { 0x04, 0x03, 0x02, 0x01 };

        Event event1( 0x1a2b3c4d, ConstSimpleData( data, sizeof(data) ) );

        HeapData serialized;
        event1.serialize( serialized );

        char expectedSeriazlied[] = {
            0x4d, 0x3c, 0x2b, 0x1a,
            0x04, 0x00, 0x00, 0x00,
            0x04, 0x03, 0x02, 0x01 };

        TS_ASSERT_EQUALS( serialized.size(), sizeof(expectedSeriazlied) );
        TS_ASSERT_SAME_DATA( serialized.data(), expectedSeriazlied, sizeof(expectedSeriazlied) );
    }

    void test_deserialize__no_payload()
    {
        char serialized[] = { 0x4d, 0x3c, 0x2b, 0x1a,
                              0x00, 0x00, 0x00, 0x00 };

        Event event1;
        event1.deserialize( ConstSimpleData( serialized, sizeof(serialized) ) );

        TS_ASSERT_EQUALS( event1.type(), 0x1a2b3c4du );
        TS_ASSERT_EQUALS( event1.payload().size(), 0U );
        TS_ASSERT_EQUALS( event1.payloadSize(), 0U );
        TS_ASSERT_EQUALS( event1.totalSize(), 8U );
    }

    void test_deserialize__with_payload()
    {
        char serialized[] = { 0x4d, 0x3c, 0x2b, 0x1a,
                              0x04, 0x00, 0x00, 0x00,
                              0x01, 0x02, 0x03, 0x04 };

        Event event1;
        event1.deserialize( ConstSimpleData( serialized, sizeof(serialized) ) );

        TS_ASSERT_EQUALS( event1.type(), 0x1a2b3c4du );
        TS_ASSERT_EQUALS( event1.payload().size(), 4U );
        TS_ASSERT_SAME_DATA( event1.payload().data(), serialized + 8, 4U );
        TS_ASSERT_EQUALS( event1.payloadSize(), 4U );
        TS_ASSERT_EQUALS( event1.totalSize(), 12U );
    }

    void test_smaller()
    {
        Event event1( 1 );
        Event event2( 2 );
        Event event3( 3 );

        TS_ASSERT( event1 < event2 );
        TS_ASSERT( event1 < event3 );
        TS_ASSERT( event2 < event3 );

        TS_ASSERT( !( event2 < event1 ) );
        TS_ASSERT( !( event3 < event1 ) );
        TS_ASSERT( !( event3 < event2 ) );
    }

    void test_assignment()
    {
        char payload[] = { 0xe1, 0xd2, 0xc3, 0xb4 };

        Event event1( 1, ConstSimpleData( payload, sizeof(payload) ) );

        Event event2;
        TS_ASSERT_EQUALS( event2.type(), (unsigned)-1 );
        TS_ASSERT_EQUALS( event2.payload().size(), 0U );

        event2 = event1;

        TS_ASSERT_EQUALS( event2.type(), 1U );
        TS_ASSERT_EQUALS( event2.payload().size(), 4U );
        TS_ASSERT_EQUALS( event2.totalSize(), 12U );
        TS_ASSERT_EQUALS( event2.payloadSize(), 4U );
        TS_ASSERT_SAME_DATA( event2.payload().data(), payload, sizeof(payload) );
    }

    void test_copy_constructor()
    {
        char payload[] = { 0xe1, 0xd2, 0xc3, 0xb4 };

        Event event1( 1, ConstSimpleData( payload, sizeof(payload) ) );

        Event event2( event1 );

        TS_ASSERT_EQUALS( event2.type(), 1U );
        TS_ASSERT_EQUALS( event2.payload().size(), 4U );
        TS_ASSERT_EQUALS( event2.totalSize(), 12U );
        TS_ASSERT_EQUALS( event2.payloadSize(), 4U );
        TS_ASSERT_SAME_DATA( event2.payload().data(), payload, sizeof(payload) );
    }
};

//
// Local Variables:
// compile-command: "make -C ../.. "
// tags-file-name: "../../TAGS"
// c-basic-offset: 4
// End:
//
