#ifndef __NFA_h__
#define __NFA_h__

#include "Map.h"
#include "Queue.h"
#include "String.h"
#include "Vector.h"

class NFA
{
public:
    struct Run
    {
        Vector<String> _letters;
        Vector<String> _states;

        bool empty() const
        {
            return _states.empty();
        }

        void clear()
        {
            _letters.clear();
            _states.clear();
        }
    };

    struct Edge
    {
        Edge( const String &source, const String &letter, const String &target ) : _source( source ), _letter( letter ), _target( target )
        {
        }

        String _source;
        String _letter;
        String _target;

        bool operator==( const Edge &other )
        {
            return ( _source == other._source ) && ( _letter == other._letter ) && ( _target == other._target );
        }
    };

    typedef Vector<Edge> Edges;

    NFA() : _edgeCount( 0 )
    {
    }

    void addLetter( const String &letter )
    {
        _alphabet.append( letter );
    }

    void addState( const String &state )
    {
        _states.append( state );
    }

    void removeState( const String &state )
    {
        _states.erase( state );
    }

    bool edgeExists( Edge edge )
    {
        if ( !_edgesBySource.exists( edge._source ) )
        {
            return false;
        }

        return _edgesBySource[edge._source].exists( edge );
    }

    void addEdge( Edge edge )
    {
        if ( !_edgesBySource.exists( edge._source ) )
        {
            Edges edges;
            _edgesBySource[edge._source] = edges;
        }

        if ( !_edgesByTarget.exists( edge._target ) )
        {
            Edges edges;
            _edgesByTarget[edge._target] = edges;
        }

        if ( !_edgesBySource[edge._source].exists( edge ) )
        {
            _edgesBySource[edge._source].append( edge );
            _edgesByTarget[edge._target].append( edge );
            ++_edgeCount;
        }
    }

    unsigned edgeCount() const
    {
        return _edgeCount;
    }

    void setInitialState( const String &initialState )
    {
        _initialState = initialState;
    }

    String getInitialState() const
    {
        return _initialState;
    }

    void markAsAccepting( const String &acceptingState )
    {
        _acceptingStates.append( acceptingState );
    }

    unsigned numberOfStates() const
    {
        return _states.size();
    }

    String getState( unsigned i )
    {
        return _states[i];
    }

    bool acceptingState( const String &state )
    {
        return _acceptingStates.exists( state );
    }

    bool existsSuccessor( const String &state, const String &letter )
    {
        if ( !_edgesBySource.exists( state ) )
        {
            return false;
        }

        for ( unsigned i = 0; i < _edgesBySource[state].size(); ++i )
        {
            if ( _edgesBySource[state][i]._letter == letter )
            {
                return true;
            }
        }

        return false;
    }

    String getSuccessor( const String &state, const String &letter )
    {
        for ( unsigned i = 0; i < _edgesBySource[state].size(); ++i )
        {
            if ( _edgesBySource[state][i]._letter == letter )
            {
                return _edgesBySource[state][i]._target;
            }
        }

        //printf( "State: %s. Letter: %s\n", state.ascii(), letter.ascii() );
        throw Error( Error::NO_SUCCESSOR_IN_NFA );
    }

    Vector<String> getAllSuccessors( const String &state, const String &letter )
    {
        UniqueVector<String> successors;

        for ( unsigned i = 0; i < _edgesBySource[state].size(); ++i )
        {
            if ( _edgesBySource[state][i]._letter == letter )
            {
                successors.append( _edgesBySource[state][i]._target );
            }
        }

        return successors;
    }

    void dump()
    {
        printf( "Dumping NFA:\n" );

        printf( "Number of letters: %u\nLetters are:", _alphabet.size() );
        for ( unsigned i = 0; i < _alphabet.size() - 1; ++i )
        {
            printf( "%s, ", _alphabet[i].ascii() );
        }
        printf( "%s\n\n", _alphabet[_alphabet.size() - 1].ascii() );

        printf( "Number of states: %u\n", _states.size() );
        for ( unsigned i = 0; i < _states.size(); ++i )
        {
            printf( "%s\n", _states[i].ascii() );
        }

        printf( "Initial state: %s\n\n", _initialState.ascii() );

        printf( "Number of accepting states: %u\n", _acceptingStates.size() );
        for ( unsigned i = 0; i < _acceptingStates.size(); ++i )
        {
            printf( "%s\n", _acceptingStates[i].ascii() );
        }

        printf( "\nNumber of edges: %u\n", _edgeCount );

        Map<String, Edges>::iterator it;

        for ( it = _edgesBySource.begin(); it != _edgesBySource.end(); ++it )
        {
            for ( unsigned i = 0; i < it->second.size(); ++i )
            {
                printf( "(%s, %s, %s)\n", it->second[i]._source.ascii(), it->second[i]._letter.ascii(), it->second[i]._target.ascii() );
            }
        }
    }

    bool emptyLanguage() const
    {
        return _acceptingStates.empty();
    }

    NFA::Run findRunInLanguage()
    {
        runBfs();

        String state = reachableAcceptingState();

        NFA::Run run;

        while ( predecessor( state ) != "" )
        {
            if ( acceptingState( predecessor( state ) ) )
            {
                run.clear();
            }
            else
            {
                run._letters.insertHead( letterLeadingToState( state ) );
                run._states.insertHead( state );
            }

            state = predecessor( state );
        }

        if ( acceptingState( predecessor( state ) ) )
        {
            run.clear();
        }

        run._states.insertHead( state );

        return run;
    }

    Vector<String> getAcceptingStates() const
    {
        return _acceptingStates;
    }

    void eliminateNonReachables()
    {
        runBfs();
        removeNonReachableStates();
    }

    Edges successorEdges( const String &state )
    {
        if ( _edgesBySource.exists( state ) )
        {
            return _edgesBySource[state];
        }

        Edges empty;
        return empty;
    }

    Edges predecessorEdges( const String &state )
    {
        if ( _edgesByTarget.exists( state ) )
        {
            return _edgesByTarget[state];
        }

        Edges empty;
        return empty;
    }

    UniqueVector<String> allPredecessors( const String &state )
    {
        UniqueVector<String> predecessors;

        if ( _edgesByTarget.exists( state ) )
        {
            for ( unsigned i = 0; i < _edgesByTarget[state].size(); ++i )
            {
                predecessors.append( _edgesByTarget[state][i]._source );
            }
        }

        return predecessors;
    }

    UniqueVector<String> allSuccessors( const String &state )
    {
        UniqueVector<String> successors;

        if ( _edgesBySource.exists( state ) )
        {
            for ( unsigned i = 0; i < _edgesBySource[state].size(); ++i )
            {
                successors.append( _edgesBySource[state][i]._target );
            }
        }

        return successors;
    }

    void clear()
    {
        _states.clear();
        _alphabet.clear();
        _edgesBySource.clear();
        _edgesByTarget.clear();
        _edgeCount = 0;
        _initialState = "ERROR";
        _acceptingStates.clear();

        _statesToVisit.clear();
        _visitedStates.clear();
    }

    void removeEdgesOfState( const String &state )
    {
        removeEdgesFromState( state );
        removeEdgesToState( state );
    }

    void removeEdgesFromState( const String &state )
    {
        if ( _edgesBySource.exists( state ) )
        {
            _edgeCount -= _edgesBySource[state].size();

            for ( unsigned i = 0; i < _edgesBySource[state].size(); ++i )
            {
                _edgesByTarget[_edgesBySource[state][i]._target].erase( _edgesBySource[state][i] );
            }

            _edgesBySource.erase( state );
        }
    }

    void removeEdgesToState( const String &state )
    {
        if ( _edgesByTarget.exists( state ) )
        {
            _edgeCount -= _edgesByTarget[state].size();

            for ( unsigned i = 0; i < _edgesByTarget[state].size(); ++i )
            {
                _edgesBySource[_edgesByTarget[state][i]._source].erase( _edgesByTarget[state][i] );
            }

            _edgesByTarget.erase( state );
        }
    }

private:
    Vector<String> _states;
    Vector<String> _alphabet;
    Map<String, Edges> _edgesBySource;
    Map<String, Edges> _edgesByTarget;

    String _initialState;
    Vector<String> _acceptingStates;

    Queue<String> _statesToVisit;
    Map<String, String> _visitedStates;
    unsigned _edgeCount;

    void runBfs()
    {
        initBfs();

        while ( haveStatesToVisit() )
        {
            visitSingleState();
        }
    }

    void initBfs()
    {
        _statesToVisit.clear();
        _visitedStates.clear();

        _statesToVisit.push( _initialState );
        _visitedStates[_initialState] = "";
    }

    bool haveStatesToVisit() const
    {
        return !_statesToVisit.empty();
    }

    void visitSingleState()
    {
        String state = _statesToVisit.peak();
        _statesToVisit.pop();

        if ( _edgesBySource.exists( state ) )
        {
            for ( unsigned i = 0; i < _edgesBySource[state].size(); ++i )
            {
                if ( !_visitedStates.exists( _edgesBySource[state][i]._target ) )
                {
                    _visitedStates[_edgesBySource[state][i]._target] = state;
                    _statesToVisit.push( _edgesBySource[state][i]._target );
                }
            }
        }
    }

    void removeNonReachableStates()
    {
        Vector<String>::iterator it = _states.begin();

        while ( it != _states.end() )
        {
            if ( !_visitedStates.exists( *it ) )
            {
                removeEdgesOfState( *it );
                removeFromAcceptingStates( *it );
                _states.erase( it );
            }
            else
            {
                ++it;
            }
        }
    }

    void removeFromAcceptingStates( String &state )
    {
        if ( acceptingState( state ) )
        {
            _acceptingStates.erase( state );
        }
    }

    String reachableAcceptingState()
    {
        if ( _acceptingStates.empty() )
        {
            throw Error( Error::NO_REACHABLE_ACCEPTING_STATES );
        }

        return _acceptingStates[0];
    }

    String predecessor( String &state )
    {
        return _visitedStates[state];
    }

    String letterLeadingToState( String &state )
    {
        String predecessorState = predecessor( state );

        for ( unsigned i = 0; i < _edgesBySource[predecessorState].size(); ++i )
        {
            if ( _edgesBySource[predecessorState][i]._target == state )
            {
                return _edgesBySource[predecessorState][i]._letter;
            }
        }

        throw Error( Error::NO_REACHABLE_ACCEPTING_STATES );
    }
};

#endif // __NFA_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
