#ifndef __SafetyPatcher_h__
#define __SafetyPatcher_h__

#include "SafetyPatch.h"

class SafetyPatcher
{
public:
    static SafetyPatch patch( NFA &automaton )
    {
        return SafetyPatcher( automaton ).run();
    }

private:
    NFA &_automaton;
    SafetyPatch _patch;

    unsigned _deadlockedPredecessorIndex;

    UniqueVector<String> _bad;
    UniqueVector<String> _badFrontier;
    UniqueVector<String> _predecessorsOfBad;

    SafetyPatcher( NFA &automaton ) : _automaton( automaton )
    {
        _patch.setNfa( _automaton );
    }

    SafetyPatch run()
    {
        initializeBad();
        initializePredecessors();

        printf( "Initial dump:\n" );
        dump();

        if ( initialStateIsBad() )
        {
            TS_TRACE( "Throwing" );
            throw Error( Error::RUN_UNPATCHABLE );
        }

        while ( deadlockedPredecessorExists() )
        {
            markDeadlockedPredecessorAsBad();
            dump();

            if ( initialStateIsBad() )
            {
                TS_TRACE( "Throwing" );
                throw Error( Error::RUN_UNPATCHABLE );
            }
        }

        preparePatch();

        return _patch;
    }

    void initializeBad()
    {
        _bad = _automaton.getAcceptingStates();
    }

    void initializePredecessors()
    {
        for ( unsigned i = 0; i < _bad.size(); ++i )
        {
            Vector<String> predecessors = _automaton.allPredecessors( _bad[i] );

            for ( unsigned j = 0; j < predecessors.size(); ++j )
            {
                if ( !_bad.exists( predecessors[j] ) )
                {
                    _badFrontier.append( _bad[i] );
                    _predecessorsOfBad.append( predecessors[j] );
                }
            }
        }
    }

    bool deadlockedPredecessorExists()
    {
        for ( _deadlockedPredecessorIndex = 0; _deadlockedPredecessorIndex < _predecessorsOfBad.size(); ++_deadlockedPredecessorIndex )
        {
            if ( predecessorIsDeadlocked() )
            {
                return true;
            }
        }

        return false;
    }

    bool predecessorIsDeadlocked()
    {
        Vector<String> successors = _automaton.allSuccessors( _predecessorsOfBad[_deadlockedPredecessorIndex] );

        for ( unsigned i = 0; i < successors.size(); ++i )
        {
            if ( !_bad.exists( successors[i] ) )
            {
                return false;
            }
        }

        return true;
    }

    void markDeadlockedPredecessorAsBad()
    {
        String deadlockedState = _predecessorsOfBad[_deadlockedPredecessorIndex];

        printf( "Working on deadlocked prdecessor: %s\n", deadlockedState.ascii() );

        _bad.append( deadlockedState );
        _predecessorsOfBad.erase( deadlockedState );

        if ( stateIsInTheFrontier( deadlockedState ) )
        {
            printf( "New state is in the frontier\n" );

            _badFrontier.append( deadlockedState );
            addPredecessorsToPre( deadlockedState );
        }

        removeSuccessorsFromFrontierIfNeeded( deadlockedState );
    }

    void addPredecessorsToPre( String &newBadState )
    {
        Vector<String> predecessors = _automaton.allPredecessors( newBadState );

        for ( unsigned i = 0; i < predecessors.size(); ++i )
        {
            printf( "Considering predecessor: %s ---", predecessors[i].ascii() );

            if ( !_bad.exists( predecessors[i] ) )
            {
                _predecessorsOfBad.append( predecessors[i] );
                printf( "Added!\n" );
            }
            else
            {
                printf( "Not added\n" );
            }
        }
    }

    void removeSuccessorsFromFrontierIfNeeded( String &state )
    {
        Vector<String> successors = _automaton.allSuccessors( state );

        for ( unsigned i = 0; i < successors.size(); ++i )
        {
            printf( "Removing successor from frontier (if needed): %s\n", successors[i].ascii() );
            Vector<String> predecessors = _automaton.allPredecessors( successors[i] );

            bool haveGoodPredecessors = false;

            for ( unsigned j = 0; j < predecessors.size(); ++j )
            {
                if ( !_bad.exists( predecessors[j] ) )
                {
                    haveGoodPredecessors = true;
                }
            }

            TS_TRACE( 3 );
            if ( !haveGoodPredecessors )
            {
                _badFrontier.erase( successors[i] );
            }
            TS_TRACE( 4 );
        }
    }

    bool stateIsInTheFrontier( String &badState )
    {
        Vector<String> predecessors = _automaton.allPredecessors( badState );

        for ( unsigned i = 0; i < predecessors.size(); ++i )
        {
            if ( !_bad.exists( predecessors[i] ) )
            {
                return true;
            }
        }

        return false;
    }

    void preparePatch()
    {
        for ( unsigned i = 0; i < _predecessorsOfBad.size(); ++i )
        {
            patchSingleState( _predecessorsOfBad[i] );
        }
    }

    void patchSingleState( String &predecessorState )
    {
        for ( unsigned i = 0; i < _automaton.edgeCount(); ++i )
        {
            if ( edgeLeadsToBadStates( predecessorState, _automaton.getEdge( i ) ) )
            {
                _patch.addBlockedEvent( predecessorState, letterToEvent( _automaton.getEdge( i )._letter ) );
            }
        }
    }

    bool edgeLeadsToBadStates( String &source, NFA::Edge edge )
    {
        return ( edge._source == source ) && ( _bad.exists( edge._target ) );
    }

    static Event letterToEvent( const String &letter )
    {
        return Event( atoi( letter.ascii() ) );
    }

    void dump()
    {
        printf( "Bad states:\n ");
        for ( unsigned i = 0; i < _bad.size(); ++i )
        {
            printf( "\t%s\n", _bad[i].ascii() );
        }

        printf( "Frontier states:\n ");
        for ( unsigned i = 0; i < _badFrontier.size(); ++i )
        {
            printf( "\t%s\n", _badFrontier[i].ascii() );
        }

        printf( "Predecessor states:\n ");
        for ( unsigned i = 0; i < _predecessorsOfBad.size(); ++i )
        {
            printf( "\t%s\n", _predecessorsOfBad[i].ascii() );
        }
    }

    bool initialStateIsBad()
    {
        return _bad.exists( _automaton.getInitialState() );
    }
};

#endif // __SafetyPatcher_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
