#ifndef __HttpGetHandler_h__
#define __HttpGetHandler_h__

#include "BThread.h"
#include "File.h"
#include "IEvent.h"
#include "ServerConfiguration.h"

class HttpGetHandler : public BThread
{
public:
    void entryPoint()
    {
        while ( true )
        {
            waitForHttpGet();
            extractFileName();

            if ( fileDoesntExist() )
            {
                printf( "File doesnt exist\n" );
                fflush( 0 );
                sendErrorMessage();
            }
            else if ( directoryNotEndingWithSlash() )
            {
                printf( "Directory not ending with slash\n" );
                fflush( 0 );
                sendFoundMessage();
            }
            else
            {
                considerDefaultFileName();
                sendResponse();
                endSession();
                doneHandlingRequest();
            }
        }
    }

private:
    enum {
        MAX_CHUNK_SIZE = 4096
    };

    String _fileName;
    String _fileNameFullPath;

    void waitForHttpGet()
    {
        clearEventVectors();

        _watched.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked );
    }

    void extractFileName()
    {
        String requestString( lastEvent().payload().asChar(), lastEvent().payloadSize() );
        char delimiter = ' ';
        _fileName = requestString.tokenize( &delimiter )[1];
        fixSpaceEncoding();
        _fileNameFullPath = String( ServerConfiguration::LOCAL_DIRECTORY ) + _fileName;

        printf( "\t    GET: %s\n", _fileNameFullPath.ascii() );
        fflush( 0 );
    }

    void fixSpaceEncoding()
    {
        static const char WRONG_ENCODING[] = "%20";
        static const char CORRECT_ENCODING[] = " ";

        while ( _fileName.contains( WRONG_ENCODING ) )
        {
            _fileName.replace( WRONG_ENCODING, CORRECT_ENCODING );
        }
    }

    bool directoryNotEndingWithSlash()
    {
        return File::directory( _fileNameFullPath ) && ( _fileNameFullPath[_fileNameFullPath.length() - 1] != '/' );
    }

    void considerDefaultFileName()
    {
        if ( File::directory( _fileNameFullPath ) )
        {
            appendDefaultSuffix();
        }
    }

    void appendDefaultSuffix()
    {
        static char DEFAULT_FILE_NAME[] = "index.html";
        _fileNameFullPath += DEFAULT_FILE_NAME;
    }

    bool fileDoesntExist()
    {
        return !File::exists( _fileNameFullPath );
    }

    String contentType()
    {
        String type;

        if ( _fileNameFullPath.contains( ".html" ) )
        {
            type = "text/html; charset=iso-8859-1";
        }
        else if ( _fileNameFullPath.contains( ".css" ) )
        {
            type = "text/css; charset=iso-8859-1";
        }
        else if ( _fileNameFullPath.contains( ".js" ) )
        {
            type = "application/x-javascript";
        }
        else
        {
            type = "text/plain; charset=iso-8859-1";
        }

        return Stringf( "Content-Type: %s\r\n", type.ascii() );
    }

    void sendResponse()
    {
        sendHeader();
        sendContent();
    }

    void sendHeader()
    {
        Stringf headerString( "%s%s%s%s%u%s",
                              "HTTP/1.1 200 Ok\r\n",
                              "Server: BP Server\r\n",
                              contentType().ascii(),
                              "Content-Length: ",
                              File::getSize( _fileNameFullPath ),
                              "\r\nConnection: close\r\n\r\n" );

        HeapData header = ConstSimpleData( headerString.ascii(), headerString.length() );

        clearEventVectors();
        _requested.append( Event( IEvent::DATA_FROM_HTTP_TO_TCP, header ) );
        _blocked.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked );
    }

    void sendContent()
    {
        File inputFile( _fileNameFullPath );
        inputFile.open( IFile::MODE_READ );

        unsigned indexInFile = 0;

        while ( indexInFile < File::getSize( _fileNameFullPath ) )
        {
            sendChunk( inputFile );
            indexInFile += MAX_CHUNK_SIZE;
        }
    }

    void sendChunk( IFile &file )
    {
        HeapData content;
        file.read( content, MAX_CHUNK_SIZE );

        clearEventVectors();
        _requested.append( Event( IEvent::DATA_FROM_HTTP_TO_TCP, content ) );
        _blocked.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked );
    }

    void sendErrorMessage()
    {
        clearEventVectors();
        _requested.append( IEvent::HTTP_GENERATE_ERROR );
        _blocked.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked );
    }

    void sendFoundMessage()
    {
        clearEventVectors();
        _requested.append( Event( IEvent::HTTP_GENERATE_FOUND, ConstSimpleData( _fileName.ascii(), _fileName.length() ) ) );
        _blocked.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked );
    }

    void endSession()
    {
        clearEventVectors();
        _requested.append( IEvent::HTTP_END_SESSION );
        _blocked.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked );
    }

    void doneHandlingRequest()
    {
        clearEventVectors();
        _requested.append( IEvent::HTTP_DONE_HANDLING_REQUEST );
        _blocked.append( IEvent::HTTP_GET_REQUEST );
        bSync( _requested, _watched, _blocked );
    }
};

#endif // __HttpGetHandler_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
