#ifndef __SocketManager_h__
#define __SocketManager_h__

#include "BThread.h"
#include "Event.h"
#include "IEvent.h"
#include "Queue.h"
#include "SelectableSocket.h"
#include "Selector.h"
#include "ServerConfiguration.h"
#include "ServerSocket.h"

class SocketManager : public BThread,
                      public IServerSocket::ISocketAccepter,
                      public ISelectableSocket::ISocketDataHandler
{
public:
    SocketManager() : _serverSocket( _selector, ServerConfiguration::LISTENING_PORT, *this ), _socketToClient( NULL ), _handlingASocket( false ), _handlingARequest( false )
    {
    }

    ~SocketManager()
    {
        freeSocketIfNeeded();
    }

    void entryPoint()
    {
        while ( true )
        {
            _selector.select();

            while ( busyWithARequest() )
            {
                handleInternalEvents();
            }
        }
    }

    void handleAcceptedSocket( ISocket &socket )
    {
        printf( "+ Incoming socket\n" );
        fflush( 0 );

        addSocketToQueue( socket );

        if ( !busyWithASocket() )
        {
            startHandlingASocket();
        }
    }

    void startHandlingASocket()
    {
        _handlingASocket = true;
        createSocket( *_socketsToHandle.peak() );
        _socketsToHandle.pop();
    }

    void handleData( const ConstSimpleData &data )
    {
        printf( "\t>> Incoming data request\n" );
        fflush( 0 );

        _handlingARequest = true;

        clearEventVectors();
        _requested.append( Event( IEvent::HTTP_DATA_RECEIVED, data ) );
        bSync( _requested, _watched, _blocked );
    }

    void handleInternalEvents()
    {
        waitForAnyEvent();

        switch ( lastEvent().type() )
        {
        case IEvent::SEND_DATA:
            sendPayload();
            break;
        case IEvent::HTTP_END_SESSION:
            _socketToClient->closeSocket();
            freeSocketIfNeeded();
            break;
        case IEvent::HTTP_DONE_HANDLING_REQUEST:
            printf( "\t>> Done handling data request\n" );
            fflush( 0 );

            _handlingASocket = false;
            _handlingARequest = false;
            if ( havePendingSockets() )
            {
                startHandlingASocket();
            }
            break;
        default:
            break;
        }
    }

private:
    Selector _selector;
    ServerSocket _serverSocket;
    ISelectableSocket *_socketToClient;
    Queue<ISocket *> _socketsToHandle;
    bool _handlingASocket;
    bool _handlingARequest;

    void freeSocketIfNeeded()
    {
        if ( _socketToClient )
        {
            delete _socketToClient;
            _socketToClient = NULL;
        }
    }

    void sendPayload()
    {
        ((ISocket &)*_socketToClient).send( lastEvent().payload() );
    }

    void waitForAnyEvent()
    {
        clearEventVectors();
        _watched = IEvent::allEvents();

        bSync( _requested, _watched, _blocked );
    }

    void createSocket( ISocket &socket )
    {
        _socketToClient = new SelectableSocket( socket, _selector, *this );
    }

    bool busyWithASocket() const
    {
        return _handlingASocket;
    }

    bool busyWithARequest() const
    {
        return _handlingARequest;
    }

    bool havePendingSockets() const
    {
        return !_socketsToHandle.empty();
    }

    void addSocketToQueue( ISocket &socket )
    {
        _socketsToHandle.push( &socket );
    }
};

#endif // __SocketManager_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
